@@ -18,8 +18,8 @@ use std::collections::HashMap;
 use std::collections::hash_map::Entry;
 use std::fmt::{Display, Formatter, Error as FmtError};
 use std::mem::replace;
-use crc::vorbis_crc32_update;
-use Packet;
+use crate::crc::vorbis_crc32_update;
+use crate::Packet;
 use std::io::Seek;
 
 /// Error that can be raised when decoding an Ogg transport.
@@ -33,27 +33,33 @@ pub enum OggReadError {
 	InvalidStreamStructVer(u8),
 	/// Mismatch of the hash value with (expected, calculated) value.
 	HashMismatch(u32, u32),
-	/// I/O error occured.
+	/// I/O error occurred.
 	ReadError(io::Error),
 	/// Some constraint required by the spec was not met.
 	InvalidData,
 }
 
-impl error::Error for OggReadError {
-	fn description(&self) -> &str {
-		match self {
-			&OggReadError::NoCapturePatternFound => "No Ogg capture pattern found",
-			&OggReadError::InvalidStreamStructVer(_) =>
+impl OggReadError {
+	fn description_str(&self) -> &str {
+		match *self {
+			OggReadError::NoCapturePatternFound => "No Ogg capture pattern found",
+			OggReadError::InvalidStreamStructVer(_) =>
 				"A non zero stream structure version was passed",
-			&OggReadError::HashMismatch(_, _) => "CRC32 hash mismatch",
-			&OggReadError::ReadError(_) => "I/O error",
-			&OggReadError::InvalidData => "Constraint violated",
+			OggReadError::HashMismatch(_, _) => "CRC32 hash mismatch",
+			OggReadError::ReadError(_) => "I/O error",
+			OggReadError::InvalidData => "Constraint violated",
 		}
 	}
+}
 
-	fn cause(&self) -> Option<&error::Error> {
-		match self {
-			&OggReadError::ReadError(ref err) => Some(err as &error::Error),
+impl error::Error for OggReadError {
+	fn description(&self) -> &str {
+		self.description_str()
+	}
+
+	fn cause(&self) -> Option<&dyn error::Error> {
+		match *self {
+			OggReadError::ReadError(ref err) => Some(err as &dyn error::Error),
 			_ => None
 		}
 	}
@@ -61,7 +67,7 @@ impl error::Error for OggReadError {
 
 impl Display for OggReadError {
 	fn fmt(&self, fmt :&mut Formatter) -> Result<(), FmtError> {
-		write!(fmt, "{}", error::Error::description(self))
+		write!(fmt, "{}", Self::description_str(self))
 	}
 }
 
@@ -73,7 +79,7 @@ impl From<io::Error> for OggReadError {
 
 /// Containing information about an OGG page that is shared between multiple places
 struct PageBaseInfo {
-	/// `true`: the first packet is continued from the page before. `false`: if its a "fresh" one
+	/// `true`: the first packet is continued from the page before. `false`: if it's a "fresh" one
 	starts_with_continued :bool,
 	/// `true` if this page is the first one in the logical bitstream
 	first_page :bool,
@@ -152,7 +158,7 @@ impl OggPage {
 /**
 Helper struct for parsing pages
 
-Its created using the `new` function and then its fed more data via the `parse_segments`
+It's created using the `new` function and then it's fed more data via the `parse_segments`
 and `parse_packet_data` functions, each called exactly once and in that precise order.
 
 Then later code uses the `OggPage` returned by the `parse_packet_data` function.
@@ -166,7 +172,7 @@ pub struct PageParser {
 	checksum :u32,
 	header_buf: [u8; 27],
 	/// Number of packet ending segments
-	packet_count :u16, // Gets populated gafter segments have been parsed
+	packet_count :u16, // Gets populated after segments have been parsed
 	/// after segments have been parsed, this contains the segments buffer,
 	/// after the packet data have been read, this contains the packets buffer.
 	segments_or_packets_buf :Vec<u8>,
@@ -187,29 +193,29 @@ impl PageParser {
 	pub fn new(header_buf :[u8; 27]) -> Result<(PageParser, usize), OggReadError> {
 		let mut header_rdr = Cursor::new(header_buf);
 		header_rdr.set_position(4);
-		let stream_structure_version = try!(header_rdr.read_u8());
+		let stream_structure_version = tri!(header_rdr.read_u8());
 		if stream_structure_version != 0 {
-			try!(Err(OggReadError::InvalidStreamStructVer(stream_structure_version)));
+			tri!(Err(OggReadError::InvalidStreamStructVer(stream_structure_version)));
 		}
 		let header_type_flag = header_rdr.read_u8().unwrap();
-		let stream_serial;
+		let absgp = header_rdr.read_u64::<LittleEndian>().unwrap();
+		let stream_serial = header_rdr.read_u32::<LittleEndian>().unwrap();
+		let sequence_num = header_rdr.read_u32::<LittleEndian>().unwrap();
+		let checksum = header_rdr.read_u32::<LittleEndian>().unwrap();
 
 		Ok((PageParser {
 			bi : PageBaseInfo {
 				starts_with_continued : header_type_flag & 0x01u8 != 0,
 				first_page : header_type_flag & 0x02u8 != 0,
 				last_page : header_type_flag & 0x04u8 != 0,
-				absgp : header_rdr.read_u64::<LittleEndian>().unwrap(),
-				sequence_num : {
-					stream_serial = header_rdr.read_u32::<LittleEndian>().unwrap();
-					header_rdr.read_u32::<LittleEndian>().unwrap()
-				},
+				absgp,
+				sequence_num,
 				packet_positions : Vec::new(),
 				ends_with_continued : false,
 			},
-			stream_serial : stream_serial,
-			checksum : header_rdr.read_u32::<LittleEndian>().unwrap(),
-			header_buf : header_buf,
+			stream_serial,
+			checksum,
+			header_buf,
 			packet_count : 0,
 			segments_or_packets_buf :Vec::new(),
 		},
@@ -244,7 +250,7 @@ impl PageParser {
 		let mut cur_packet_offs :u16 = 0;
 
 		// Second run: get the offsets of the packets
-		// Not that we need it right now, but its much more fun this way, am I right
+		// Not that we need it right now, but it's much more fun this way, am I right
 		for val in &segments_buf {
 			cur_packet_siz += *val as u16;
 			if *val < 255 {
@@ -287,7 +293,7 @@ impl PageParser {
 			// This allows random input from fuzzers reach decoding code that's actually interesting,
 			// instead of being rejected early due to checksum mismatch.
 			if !cfg!(fuzzing) {
-				try!(Err(OggReadError::HashMismatch(self.checksum, hash_calculated)));
+				tri!(Err(OggReadError::HashMismatch(self.checksum, hash_calculated)));
 			}
 		}
 		self.segments_or_packets_buf = packet_data;
@@ -325,7 +331,7 @@ pub struct BasePacketReader {
 	/// None if there is no such stream and one has to read a new page.
 	stream_with_stuff :Option<u32>,
 
-	// Bool that is set to true when a seek of the stream has occured.
+	// Bool that is set to true when a seek of the stream has occurred.
 	// This helps validator code to decide whether to accept certain strange data.
 	has_seeked :bool,
 }
@@ -376,7 +382,7 @@ impl BasePacketReader {
 			cont
 		} else {
 			let mut cont :Vec<u8> = Vec::with_capacity(len as usize);
-			// TODO The copy below is totally unneccessary. It is only needed so that we don't have to carry around the old Vec's.
+			// TODO The copy below is totally unnecessary. It is only needed so that we don't have to carry around the old Vec's.
 			// TODO get something like the shared_slice crate for RefCells, so that we can also have mutable data, shared through
 			// slices.
 			let cont_slice :&[u8] = &pg_info.page_body[offs as usize .. (offs + len) as usize];
@@ -421,11 +427,11 @@ impl BasePacketReader {
 			Entry::Occupied(mut o) => {
 				let inf = o.get_mut();
 				if pg_prs.bi.first_page {
-					try!(Err(OggReadError::InvalidData));
+					tri!(Err(OggReadError::InvalidData));
 				}
 				if pg_prs.bi.starts_with_continued != inf.bi.ends_with_continued {
 					if !self.has_seeked {
-						try!(Err(OggReadError::InvalidData));
+						tri!(Err(OggReadError::InvalidData));
 					} else {
 						// If we have seeked, we are more tolerant here,
 						// and just drop the continued packet's content.
@@ -447,14 +453,14 @@ impl BasePacketReader {
 					}
 				} else if pg_prs.bi.starts_with_continued {
 					// Remember the packet at the end so that it can be glued together once
-					// we encounter the next segment with length < 255 (doesnt have to be in this page)
+					// we encounter the next segment with length < 255 (doesn't have to be in this page)
 					let (offs, len) = inf.bi.packet_positions[inf.packet_idx as usize];
 					if len as usize != inf.page_body.len() {
 						let mut tmp = Vec::with_capacity(len as usize);
 						tmp.write_all(&inf.page_body[offs as usize .. (offs + len) as usize]).unwrap();
 						inf.last_overlap_pck.push(tmp);
 					} else {
-						// Little optimisation: don't copy if not neccessary
+						// Little optimisation: don't copy if not necessary
 						inf.last_overlap_pck.push(replace(&mut inf.page_body, vec![0;0]));
 					}
 
@@ -467,7 +473,7 @@ impl BasePacketReader {
 				if !self.has_seeked {
 					if !pg_prs.bi.first_page || pg_prs.bi.starts_with_continued {
 						// If we haven't seeked, this is an error.
-						try!(Err(OggReadError::InvalidData));
+						tri!(Err(OggReadError::InvalidData));
 					}
 				} else {
 					if !pg_prs.bi.first_page {
@@ -475,7 +481,7 @@ impl BasePacketReader {
 					}
 					if pg_prs.bi.starts_with_continued {
 						// Ignore the continued packet's content.
-						// This is a normal occurence if we have just seeked.
+						// This is a normal occurrence if we have just seeked.
 						pg_prs.bi.packet_positions.remove(0);
 						if pg_prs.packet_count != 0 {
 							// Decrease packet count by one. Normal case.
@@ -558,16 +564,16 @@ impl UntilPageHeaderReader {
 		}
 	}
 	/// Returns Some(off), where off is the offset of the last byte
-	/// of the capture pattern if its found, None if the capture pattern
+	/// of the capture pattern if it's found, None if the capture pattern
 	/// is not inside the passed slice.
 	///
 	/// Changes the capture pattern offset accordingly
 	fn check_arr(&mut self, arr :&[u8]) -> Option<usize> {
 		for (i, ch) in arr.iter().enumerate() {
 			match *ch {
-				0x4f /*'O'*/ if self.cpt_of == 0 => self.cpt_of = 1,
-				0x67 /*'g'*/ if self.cpt_of == 1 || self.cpt_of == 2 => self.cpt_of += 1,
-				0x53 /*'S'*/ if self.cpt_of == 3 => return Some(i),
+				b'O' => self.cpt_of = 1,
+				b'g' if self.cpt_of == 1 || self.cpt_of == 2 => self.cpt_of += 1,
+				b'S' if self.cpt_of == 3 => return Some(i),
 				_ => self.cpt_of = 0,
 			}
 		}
@@ -581,11 +587,11 @@ impl UntilPageHeaderReader {
 			-> Result<UntilPageHeaderResult, OggReadError> {
 		use self::UntilPageHeaderReaderMode::*;
 		use self::UntilPageHeaderResult as Res;
-		// The array's size is freely choseable, but must be > 27,
+		// The array's size is freely choosable, but must be > 27,
 		// and must well fit into an i32 (needs to be stored in SeekNeeded)
 		let mut buf :[u8; 1024] = [0; 1024];
 
-		let rd_len = try!(rdr.read(if self.read_amount < 27 {
+		let rd_len = tri!(rdr.read(if self.read_amount < 27 {
 			// This is an optimisation for the most likely case:
 			// the next page directly follows the current read position.
 			// Then it would be a waste to read more than the needed amount.
@@ -600,19 +606,17 @@ impl UntilPageHeaderReader {
 		}));
 
 		if rd_len == 0 {
-			// Reached EOF.
-			if self.read_amount == 0 {
-				// If we have read nothing yet, there is no data
-				// but ogg data, meaning the stream ends legally
-				// and without corruption.
-				return Ok(Res::Eof);
-			} else {
-				// There is most likely a corruption here.
-				// I'm not sure, but the ogg spec doesn't say that
-				// random data past the last ogg page is allowed,
-				// so we just assume its not allowed.
-				try!(Err(OggReadError::NoCapturePatternFound));
-			}
+			// Reached EOF. This means we're in one of these cases:
+			// 1. If we have read nothing yet (self.read_amount == 0),
+			//    there is no data but ogg data, meaning the stream
+			//    ends legally and without corruption.
+			// 2. If we have read something, there is corruption here.
+			//    The ogg spec doesn't say whether random data past the
+			//    last ogg page is allowed, and ogginfo complains about
+			//    that. But files with trailing garbarge can be played
+			//    back just fine by oggdec, VLC and other players, so
+			//    just ignore that to meet user expectations.
+			return Ok(Res::Eof);
 		}
 		self.read_amount += rd_len;
 
@@ -624,20 +628,20 @@ impl UntilPageHeaderReader {
 		if self.read_amount > read_amount_max {
 			// Exhaustive searching for the capture pattern
 			// has returned no ogg capture pattern.
-			try!(Err(OggReadError::NoCapturePatternFound));
+			tri!(Err(OggReadError::NoCapturePatternFound));
 		}
 
 		let rd_buf = &buf[0 .. rd_len];
 
 		use std::cmp::min;
-		let (off, needed) = match self.mode.clone() {
+		let (off, needed) = match self.mode {
 			Searching => match self.check_arr(rd_buf) {
 				// Capture pattern found
 				Some(off) => {
-					self.ret_buf[0] = 0x4f; // 'O'
-					self.ret_buf[1] = 0x67; // 'g'
-					self.ret_buf[2] = 0x67; // 'g'
-					self.ret_buf[3] = 0x53; // 'S' (Not actually needed)
+					self.ret_buf[0] = b'O';
+					self.ret_buf[1] = b'g';
+					self.ret_buf[2] = b'g';
+					self.ret_buf[3] = b'S'; // (Not actually needed)
 					(off, 24)
 				},
 				// Nothing found
@@ -655,6 +659,11 @@ impl UntilPageHeaderReader {
 		let start_fill = 27 - needed;
 		(&mut self.ret_buf[start_fill .. copy_amount + start_fill])
 				.copy_from_slice(&fnd_buf[0 .. copy_amount]);
+		// Comparison chain operation via cmp can be slower,
+		// and also requires an import. It's a questionable idea
+		// to suggest it. TODO: once open ranges are stable in
+		// match, use them.
+		#[allow(clippy::comparison_chain)]
 		if fnd_buf.len() == needed {
 			// Capture pattern found!
 			self.mode = Found;
@@ -677,10 +686,10 @@ impl UntilPageHeaderReader {
 			-> Result<UntilPageHeaderResult, OggReadError> {
 		use self::UntilPageHeaderReaderMode::*;
 		use self::UntilPageHeaderResult as Res;
-		match self.mode.clone() {
+		match self.mode {
 			Searching | FoundWithNeeded(_) => Ok(Res::ReadNeeded),
 			SeekNeeded(offs) => {
-				try!(skr.seek(SeekFrom::Current(offs as i64)));
+				tri!(skr.seek(SeekFrom::Current(offs as i64)));
 				self.mode = Found;
 				Ok(Res::Found)
 			},
@@ -710,12 +719,14 @@ pub struct PacketReader<T :io::Read + io::Seek> {
 	rdr :T,
 
 	base_pck_rdr :BasePacketReader,
+
+	read_some_pg :bool
 }
 
 impl<T :io::Read + io::Seek> PacketReader<T> {
 	/// Constructs a new `PacketReader` with a given `Read`.
 	pub fn new(rdr :T) -> PacketReader<T> {
-		PacketReader { rdr: rdr, base_pck_rdr : BasePacketReader::new() }
+		PacketReader { rdr, base_pck_rdr : BasePacketReader::new(), read_some_pg : false }
 	}
 	/// Returns the wrapped reader, consuming the `PacketReader`.
 	pub fn into_inner(self) -> T {
@@ -732,9 +743,9 @@ impl<T :io::Read + io::Seek> PacketReader<T> {
 			if let Some(pck) = self.base_pck_rdr.read_packet() {
 				return Ok(Some(pck));
 			}
-			let page = try!(self.read_ogg_page());
+			let page = tri!(self.read_ogg_page());
 			match page {
-				Some(page) => try!(self.base_pck_rdr.push_page(page)),
+				Some(page) => tri!(self.base_pck_rdr.push_page(page)),
 				None => return Ok(None),
 			}
 		}
@@ -745,9 +756,9 @@ impl<T :io::Read + io::Seek> PacketReader<T> {
 	/// returns an Err(_) if the physical stream has ended.
 	/// This function is useful if you expect a new packet to come.
 	pub fn read_packet_expected(&mut self) -> Result<Packet, OggReadError> {
-		match try!(self.read_packet()) {
+		match tri!(self.read_packet()) {
 			Some(p) => Ok(p),
-			None => try!(Err(Error::new(ErrorKind::UnexpectedEof,
+			None => tri!(Err(Error::new(ErrorKind::UnexpectedEof,
 				"Expected ogg packet but found end of physical stream"))),
 		}
 	}
@@ -763,13 +774,18 @@ impl<T :io::Read + io::Seek> PacketReader<T> {
 	fn read_until_pg_header(&mut self) -> Result<Option<[u8; 27]>, OggReadError> {
 		let mut r = UntilPageHeaderReader::new();
 		use self::UntilPageHeaderResult::*;
-		let mut res = try!(r.do_read(&mut self.rdr));
+		let mut res = tri!(r.do_read(&mut self.rdr));
 		loop {
 			res = match res {
 				Eof => return Ok(None),
-				Found => break,
-				ReadNeeded => try!(r.do_read(&mut self.rdr)),
-				SeekNeeded => try!(r.do_seek(&mut self.rdr))
+				Found => {
+					// Keep track whether a page was read to distinguish non-Ogg
+					// files from Ogg files with trailing junk at read_ogg_page.
+					self.read_some_pg = true;
+					break
+				},
+				ReadNeeded => tri!(r.do_read(&mut self.rdr)),
+				SeekNeeded => tri!(r.do_seek(&mut self.rdr))
 			}
 		}
 		Ok(Some(r.into_header()))
@@ -781,21 +797,22 @@ impl<T :io::Read + io::Seek> PacketReader<T> {
 	/// is at the current reader position.
 	/// Instead it searches until it finds the capture pattern.
 	fn read_ogg_page(&mut self) -> Result<Option<OggPage>, OggReadError> {
-		let header_buf :[u8; 27] = match try!(self.read_until_pg_header()) {
+		let header_buf :[u8; 27] = match tri!(self.read_until_pg_header()) {
 			Some(s) => s,
-			None => return Ok(None)
+			None if self.read_some_pg => return Ok(None),
+			None => return Err(OggReadError::NoCapturePatternFound)
 		};
-		let (mut pg_prs, page_segments) = try!(PageParser::new(header_buf));
+		let (mut pg_prs, page_segments) = tri!(PageParser::new(header_buf));
 
 		let mut segments_buf = vec![0; page_segments]; // TODO fix this, we initialize memory for NOTHING!!! Out of some reason, this is seen as "unsafe" by rustc.
-		try!(self.rdr.read_exact(&mut segments_buf));
+		tri!(self.rdr.read_exact(&mut segments_buf));
 
 		let page_siz = pg_prs.parse_segments(segments_buf);
 
 		let mut packet_data = vec![0; page_siz as usize];
-		try!(self.rdr.read_exact(&mut packet_data));
+		tri!(self.rdr.read_exact(&mut packet_data));
 
-		Ok(Some(try!(pg_prs.parse_packet_data(packet_data))))
+		Ok(Some(tri!(pg_prs.parse_packet_data(packet_data))))
 	}
 
 	/// Seeks the underlying reader
@@ -805,7 +822,7 @@ impl<T :io::Read + io::Seek> PacketReader<T> {
 	///
 	/// This also flushes all the unread packets in the queue.
 	pub fn seek_bytes(&mut self, pos :SeekFrom) -> Result<u64, Error> {
-		let r = try!(self.rdr.seek(pos));
+		let r = tri!(self.rdr.seek(pos));
 		// Reset the internal state
 		self.base_pck_rdr.update_after_seek();
 		return Ok(r);
@@ -819,12 +836,12 @@ impl<T :io::Read + io::Seek> PacketReader<T> {
 	/// at the last page that comes before such a page and has a packet start.
 	///
 	/// The passed `stream_serial` parameter controls the stream
-	/// serial number to filter our search for. If its `None`, no
+	/// serial number to filter our search for. If it's `None`, no
 	/// filtering is applied, but if it is `Some(n)`, we filter for
 	/// streams with the serial number `n`.
 	/// Note that the `None` case is only intended for streams
 	/// where only one logical stream exists, the seek may misbehave
-	/// if `Ǹone` gets passed when multiple streams exist.
+	/// if `None` gets passed when multiple streams exist.
 	///
 	/// The returned bool indicates whether the seek was successful.
 	pub fn seek_absgp(&mut self, stream_serial :Option<u32>,
@@ -832,14 +849,14 @@ impl<T :io::Read + io::Seek> PacketReader<T> {
 		macro_rules! found {
 			($pos:expr) => {{
 				// println!("found: {}", $pos);
-				try!(self.rdr.seek(SeekFrom::Start($pos)));
+				tri!(self.rdr.seek(SeekFrom::Start($pos)));
 				self.base_pck_rdr.update_after_seek();
 				return Ok(true);
 			}};
 		}
 		macro_rules! bt {
 			($e:expr) => {{
-				match try!($e) {
+				match tri!($e) {
 					Some(s) => s,
 					None => return Ok(false),
 				}
@@ -885,7 +902,7 @@ impl<T :io::Read + io::Seek> PacketReader<T> {
 				let mut pg;
 				let mut continued_pck_start = None;
 				loop {
-					pos = try!(self.rdr.seek(SeekFrom::Current(0)));
+					pos = tri!(self.rdr.seek(SeekFrom::Current(0)));
 					pg = bt!(self.read_ogg_page());
 					/*println!("absgp {} serial {} wh {} pe {} @ {}",
 						pg.0.bi.absgp, pg.0.bi.sequence_num,
@@ -927,12 +944,19 @@ impl<T :io::Read + io::Seek> PacketReader<T> {
 
 		// First, find initial "boundaries"
 		// Seek to the start of the file to get the starting boundary
-		try!(self.rdr.seek(SeekFrom::Start(0)));
+		tri!(self.rdr.seek(SeekFrom::Start(0)));
 		let (mut begin_pos, mut begin_pg) = pg_read_match_serial!();
+
+		// If the goal is the beginning, we are done.
+		if pos_goal == 0 {
+			//println!("Seeking to the beginning of the stream - skipping bisect.");
+			found!(begin_pos);
+		}
+
 		// Seek to the end of the file to get the ending boundary
 		// TODO the 200 KB is just a guessed number, any ideas
 		// to improve it?
-		try!(seek_before_end(&mut self.rdr, 200 * 1024));
+		tri!(seek_before_end(&mut self.rdr, 200 * 1024));
 		let (mut end_pos, mut end_pg) = pg_read_until_end_or_goal!(pos_goal);
 
 		// Then perform the bisection
@@ -944,7 +968,7 @@ impl<T :io::Read + io::Seek> PacketReader<T> {
 			}
 			// Perform the bisection step
 			let pos_to_seek = begin_pos + (end_pos - begin_pos) / 2;
-			try!(self.rdr.seek(SeekFrom::Start(pos_to_seek)));
+			tri!(self.rdr.seek(SeekFrom::Start(pos_to_seek)));
 			let (pos, pg) = pg_read_match_serial!();
 			/*println!("seek {} {} . {} @ {} {} . {}",
 				ab_of(&begin_pg), ab_of(&end_pg), ab_of(&pg),
@@ -958,9 +982,9 @@ impl<T :io::Read + io::Seek> PacketReader<T> {
 				let mut pos;
 				let mut pg;
 				let mut last_packet_end_pos = begin_pos;
-				try!(self.rdr.seek(SeekFrom::Start(begin_pos)));
+				tri!(self.rdr.seek(SeekFrom::Start(begin_pos)));
 				loop {
-					pos = try!(self.rdr.seek(SeekFrom::Current(0)));
+					pos = tri!(self.rdr.seek(SeekFrom::Current(0)));
 					pg = bt!(self.read_ogg_page());
 					/*println!("absgp {} pck_start {} whole_pck {} pck_end {} @ {} {}",
 						ab_of(&pg), pg.has_packet_start(), pg.has_whole_packet(),
@@ -1001,25 +1025,27 @@ impl<T :io::Read + io::Seek> PacketReader<T> {
 // util function
 fn seek_before_end<T :io::Read + io::Seek>(mut rdr :T,
 		offs :u64) -> Result<u64, OggReadError> {
-	use std::io::SeekFrom;
-	let end_pos = try!(rdr.seek(SeekFrom::End(0)));
+	let end_pos = tri!(rdr.seek(SeekFrom::End(0)));
 	let end_pos_to_seek = ::std::cmp::min(end_pos, offs);
-	return Ok(try!(rdr.seek(SeekFrom::End(-(end_pos_to_seek as i64)))));
+	return Ok(tri!(rdr.seek(SeekFrom::End(-(end_pos_to_seek as i64)))));
 }
 
 #[cfg(feature = "async")]
 /**
-Asyncronous ogg decoding
+Asynchronous ogg decoding
 */
 pub mod async_api {
-	#![allow(deprecated)]
+	use std::pin::Pin;
+	use std::task::{Context, Poll};
 
 	use super::*;
-	use tokio_io::AsyncRead;
-	use tokio_io::codec::{Decoder, FramedRead};
-	use futures::stream::Stream;
-	use futures::{Async, Poll};
+	use futures_core::{ready, Stream};
+	use futures_io::AsyncRead as FuturesAsyncRead;
+	use tokio::io::AsyncRead as TokioAsyncRead;
 	use bytes::BytesMut;
+	use pin_project::pin_project;
+	use tokio_util::codec::{Decoder, FramedRead};
+	use tokio_util::compat::{Compat, FuturesAsyncReadCompatExt};
 
 	enum PageDecodeState {
 		Head,
@@ -1031,10 +1057,10 @@ pub mod async_api {
 	impl PageDecodeState {
 		fn needed_size(&self) -> usize {
 			match self {
-				&PageDecodeState::Head => 27,
-				&PageDecodeState::Segments(_, s) => s,
-				&PageDecodeState::PacketData(_, s) => s,
-				&PageDecodeState::InUpdate => panic!("invalid state"),
+				PageDecodeState::Head => 27,
+				PageDecodeState::Segments(_, s) => *s,
+				PageDecodeState::PacketData(_, s) => *s,
+				PageDecodeState::InUpdate => panic!("invalid state"),
 			}
 		}
 	}
@@ -1075,7 +1101,7 @@ pub mod async_api {
 						// TODO once we have const generics, the copy below can be done
 						// much nicer, maybe with a new into_array fn on Vec's
 						hdr_buf.copy_from_slice(&consumed_buf);
-						let tup = try!(PageParser::new(hdr_buf));
+						let tup = tri!(PageParser::new(hdr_buf));
 						Segments(tup.0, tup.1)
 					},
 					Segments(mut pg_prs, _) => {
@@ -1083,7 +1109,7 @@ pub mod async_api {
 						PacketData(pg_prs, new_needed_len)
 					},
 					PacketData(pg_prs, _) => {
-						ret = Some(try!(pg_prs.parse_packet_data(consumed_buf)));
+						ret = Some(tri!(pg_prs.parse_packet_data(consumed_buf)));
 						Head
 					},
 					InUpdate => panic!("invalid state"),
@@ -1104,12 +1130,19 @@ pub mod async_api {
 	/**
 	Async packet reading functionality.
 	*/
-	pub struct PacketReader<T> where T :AsyncRead {
+	#[pin_project]
+	pub struct PacketReader<T> where T :TokioAsyncRead {
 		base_pck_rdr :BasePacketReader,
+		#[pin]
 		pg_rd :FramedRead<T, PageDecoder>,
 	}
 
-	impl<T :AsyncRead> PacketReader<T> {
+	impl<T :TokioAsyncRead> PacketReader<T> {
+		/// Wraps the specified Tokio runtime `AsyncRead` into an Ogg packet
+		/// reader.
+		///
+		/// This is the recommended constructor when using the Tokio runtime
+		/// types.
 		pub fn new(inner :T) -> Self {
 			PacketReader {
 				base_pck_rdr : BasePacketReader::new(),
@@ -1118,23 +1151,41 @@ pub mod async_api {
 		}
 	}
 
-	impl<T :AsyncRead> Stream for PacketReader<T> {
-		type Item = Packet;
-		type Error = OggReadError;
+	impl<T :FuturesAsyncRead> PacketReader<Compat<T>> {
+		/// Wraps the specified futures_io `AsyncRead` into an Ogg packet
+		/// reader.
+		///
+		/// This crate uses Tokio internally, so a wrapper that may have
+		/// some performance cost will be used. Therefore, this constructor
+		/// is to be used only when dealing with `AsyncRead` implementations
+		/// from other runtimes, and implementing a Tokio `AsyncRead`
+		/// compatibility layer oneself is not desired.
+		pub fn new_compat(inner :T) -> Self {
+			Self::new(inner.compat())
+		}
+	}
 
-		fn poll(&mut self) -> Poll<Option<Packet>, OggReadError> {
+	impl<T :TokioAsyncRead> Stream for PacketReader<T> {
+		type Item = Result<Packet, OggReadError>;
+
+		fn poll_next(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
+			let mut this = self.project();
 			// Read pages until we got a valid entire packet
 			// (packets may span multiple pages, so reading one page
 			// doesn't always suffice to give us a valid packet)
 			loop {
-				if let Some(pck) = self.base_pck_rdr.read_packet() {
-					return Ok(Async::Ready(Some(pck)));
-				}
-				let page = try_ready!(self.pg_rd.poll());
-				match page {
-					Some(page) => try!(self.base_pck_rdr.push_page(page)),
-					None => return Ok(Async::Ready(None)),
+				if let Some(pck) = this.base_pck_rdr.read_packet() {
+					return Poll::Ready(Some(Ok(pck)));
 				}
+				let page = match ready!(this.pg_rd.as_mut().poll_next(cx)) {
+					Some(Ok(page)) => page,
+					Some(Err(err)) => return Poll::Ready(Some(Err(err))),
+					None => return Poll::Ready(None),
+				};
+				match this.base_pck_rdr.push_page(page) {
+					Ok(_) => {},
+					Err(err) => return Poll::Ready(Some(Err(err))),
+				};
 			}
 		}
 	}
